#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Implementation of asyncio support.
#
# This file is part of pySerial. https://github.com/pyserial/pyserial-asyncio
# (C) 2015-2020 pySerial-team
#
# SPDX-License-Identifier:    BSD-3-Clause
"""\
Support asyncio with serial ports.

Posix platforms only, Python 3.5+ only.

Windows event loops can not wait for serial ports with the current
implementation. It should be possible to get that working though.
"""
import asyncio
import os

import serial

try:
    import termios
except ImportError:
    termios = None

__version__ = '0.6'


class SerialTransport(asyncio.Transport):
    """An asyncio transport model of a serial communication channel.

    A transport class is an abstraction of a communication channel.
    This allows protocol implementations to be developed against the
    transport abstraction without needing to know the details of the
    underlying channel, such as whether it is a pipe, a socket, or
    indeed a serial port.


    You generally won’t instantiate a transport yourself; instead, you
    will call `create_serial_connection` which will create the
    transport and try to initiate the underlying communication channel,
    calling you back when it succeeds.
    """

    def __init__(self, loop, protocol, serial_instance):
        super().__init__()
        self._loop = loop
        self._protocol = protocol
        self._serial = serial_instance
        self._closing = False
        self._protocol_paused = False
        self._max_read_size = 1024
        self._write_buffer = []
        self._set_write_buffer_limits()
        self._has_reader = False
        self._has_writer = False
        self._poll_wait_time = 0.0005
        self._max_out_waiting = 1024

        # XXX how to support url handlers too

        # Asynchronous I/O requires non-blocking devices
        self._serial.timeout = 0
        self._serial.write_timeout = 0

        # These two callbacks will be enqueued in a FIFO queue by asyncio
        loop.call_soon(protocol.connection_made, self)
        loop.call_soon(self._ensure_reader)

    @property
    def loop(self):
        """The asyncio event loop used by this SerialTransport."""
        return self._loop

    @property
    def serial(self):
        """The underlying Serial instance.

        Equivalent to get_extra_info("serial")
        """
        return self._serial

    def get_extra_info(self, name, default=None):
        """Get optional transport information.

        Currently only "serial" is available.
        """
        if name == "serial":
            return self._serial
        return default

    def __repr__(self):
        return '{self.__class__.__name__}({self.loop}, {self._protocol}, {self.serial})'.format(self=self)

    def is_closing(self):
        """Return True if the transport is closing or closed."""
        return self._closing

    def close(self):
        """Close the transport gracefully.

        Any buffered data will be written asynchronously. No more data
        will be received and further writes will be silently ignored.
        After all buffered data is flushed, the protocol's
        connection_lost() method will be called with None as its
        argument.
        """
        if not self._closing:
            self._close(None)

    def _read_ready(self):
        try:
            data = self._serial.read(self._max_read_size)
        except serial.SerialException as e:
            self._close(exc=e)
        else:
            if data:
                self._protocol.data_received(data)

    def write(self, data):
        """Write some data to the transport.

        This method does not block; it buffers the data and arranges
        for it to be sent out asynchronously.  Writes made after the
        transport has been closed will be ignored."""
        if self._closing:
            return

        if self.get_write_buffer_size() == 0:
            self._write_buffer.append(data)
            self._ensure_writer()
        else:
            self._write_buffer.append(data)

        self._maybe_pause_protocol()

    def can_write_eof(self):
        """Serial ports do not support the concept of end-of-file.

        Always returns False.
        """
        return False

    def pause_reading(self):
        """Pause the receiving end of the transport.

        No data will be passed to the protocol’s data_received() method
        until resume_reading() is called.
        """
        self._remove_reader()

    def resume_reading(self):
        """Resume the receiving end of the transport.

        Incoming data will be passed to the protocol's data_received()
        method until pause_reading() is called.
        """
        self._ensure_reader()

    def set_write_buffer_limits(self, high=None, low=None):
        """Set the high- and low-water limits for write flow control.

        These two values control when the protocol’s
        pause_writing()and resume_writing() methods are called. If
        specified, the low-water limit must be less than or equal to
        the high-water limit. Neither high nor low can be negative.
        """
        self._set_write_buffer_limits(high=high, low=low)
        self._maybe_pause_protocol()

    def get_write_buffer_size(self):
        """The number of bytes in the write buffer.

        This buffer is unbounded, so the result may be larger than the
        the high water mark.
        """
        return sum(map(len, self._write_buffer))

    def write_eof(self):
        raise NotImplementedError("Serial connections do not support end-of-file")

    def abort(self):
        """Close the transport immediately.

        Pending operations will not be given opportunity to complete,
        and buffered data will be lost. No more data will be received
        and further writes will be ignored.  The protocol's
        connection_lost() method will eventually be called.
        """
        self._abort(None)

    def flush(self):
        """ clears output buffer and stops any more data being written
        """
        self._remove_writer()
        self._write_buffer.clear()
        self._maybe_resume_protocol()

    def _maybe_pause_protocol(self):
        """To be called whenever the write-buffer size increases.

        Tests the current write-buffer size against the high water
        mark configured for this transport. If the high water mark is
        exceeded, the protocol is instructed to pause_writing().
        """
        if self.get_write_buffer_size() <= self._high_water:
            return
        if not self._protocol_paused:
            self._protocol_paused = True
            try:
                self._protocol.pause_writing()
            except Exception as exc:
                self._loop.call_exception_handler({
                    'message': 'protocol.pause_writing() failed',
                    'exception': exc,
                    'transport': self,
                    'protocol': self._protocol,
                })

    def _maybe_resume_protocol(self):
        """To be called whenever the write-buffer size decreases.

        Tests the current write-buffer size against the low water
        mark configured for this transport. If the write-buffer
        size is below the low water mark, the protocol is
        instructed that is can resume_writing().
        """
        if (self._protocol_paused and
                self.get_write_buffer_size() <= self._low_water):
            self._protocol_paused = False
            try:
                self._protocol.resume_writing()
            except Exception as exc:
                self._loop.call_exception_handler({
                    'message': 'protocol.resume_writing() failed',
                    'exception': exc,
                    'transport': self,
                    'protocol': self._protocol,
                })

    def _write_ready(self):
        """Asynchronously write buffered data.

        This method is called back asynchronously as a writer
        registered with the asyncio event-loop against the
        underlying file descriptor for the serial port.

        Should the write-buffer become empty if this method
        is invoked while the transport is closing, the protocol's
        connection_lost() method will be called with None as its
        argument.
        """
        data = b''.join(self._write_buffer)
        assert data, 'Write buffer should not be empty'

        self._write_buffer.clear()

        try:
            n = self._serial.write(data)
        except (BlockingIOError, InterruptedError):
            self._write_buffer.append(data)
        except serial.SerialException as exc:
            self._fatal_error(exc, 'Fatal write error on serial transport')
        else:
            if n == len(data):
                assert self._flushed()
                self._remove_writer()
                self._maybe_resume_protocol()  # May cause further writes
                # _write_ready may have been invoked by the event loop
                # after the transport was closed, as part of the ongoing
                # process of flushing buffered data. If the buffer
                # is now empty, we can close the connection
                if self._closing and self._flushed():
                    self._close()
                return

            assert 0 <= n < len(data)
            data = data[n:]
            self._write_buffer.append(data)  # Try again later
            self._maybe_resume_protocol()
            assert self._has_writer

    if os.name == "nt":
        def _poll_read(self):
            if self._has_reader and not self._closing:
                try:
                    self._has_reader = self._loop.call_later(self._poll_wait_time, self._poll_read)
                    if self.serial.in_waiting:
                        self._read_ready()
                except serial.SerialException as exc:
                    self._fatal_error(exc, 'Fatal write error on serial transport')

        def _ensure_reader(self):
            if not self._has_reader and not self._closing:
                self._has_reader = self._loop.call_later(self._poll_wait_time, self._poll_read)

        def _remove_reader(self):
            if self._has_reader:
                self._has_reader.cancel()
            self._has_reader = False

        def _poll_write(self):
            if self._has_writer and not self._closing:
                self._has_writer = self._loop.call_later(self._poll_wait_time, self._poll_write)
                if self.serial.out_waiting < self._max_out_waiting:
                    self._write_ready()

        def _ensure_writer(self):
            if not self._has_writer and not self._closing:
                self._has_writer = self._loop.call_soon(self._poll_write)

        def _remove_writer(self):
            if self._has_writer:
                self._has_writer.cancel()
            self._has_writer = False

    else:
        def _ensure_reader(self):
            if (not self._has_reader) and (not self._closing):
                self._loop.add_reader(self._serial.fileno(), self._read_ready)
                self._has_reader = True

        def _remove_reader(self):
            if self._has_reader:
                self._loop.remove_reader(self._serial.fileno())
                self._has_reader = False

        def _ensure_writer(self):
            if (not self._has_writer) and (not self._closing):
                self._loop.add_writer(self._serial.fileno(), self._write_ready)
                self._has_writer = True

        def _remove_writer(self):
            if self._has_writer:
                self._loop.remove_writer(self._serial.fileno())
                self._has_writer = False

    def _set_write_buffer_limits(self, high=None, low=None):
        """Ensure consistent write-buffer limits."""
        if high is None:
            high = 64 * 1024 if low is None else 4 * low
        if low is None:
            low = high // 4
        if not high >= low >= 0:
            raise ValueError('high (%r) must be >= low (%r) must be >= 0' %
                             (high, low))
        self._high_water = high
        self._low_water = low

    def _fatal_error(self, exc, message='Fatal error on serial transport'):
        """Report a fatal error to the event-loop and abort the transport."""
        self._loop.call_exception_handler({
            'message': message,
            'exception': exc,
            'transport': self,
            'protocol': self._protocol,
        })
        self._abort(exc)

    def _flushed(self):
        """True if the write buffer is empty, otherwise False."""
        return self.get_write_buffer_size() == 0

    def _close(self, exc=None):
        """Close the transport gracefully.

        If the write buffer is already empty, writing will be
        stopped immediately and a call to the protocol's
        connection_lost() method scheduled.

        If the write buffer is not already empty, the
        asynchronous writing will continue, and the _write_ready
        method will call this _close method again when the
        buffer has been flushed completely.
        """
        self._closing = True
        self._remove_reader()
        if self._flushed():
            self._remove_writer()
            self._loop.call_soon(self._call_connection_lost, exc)

    def _abort(self, exc):
        """Close the transport immediately.

        Pending operations will not be given opportunity to complete,
        and buffered data will be lost. No more data will be received
        and further writes will be ignored.  The protocol's
        connection_lost() method will eventually be called with the
        passed exception.
        """
        self._closing = True
        self._remove_reader()
        self._remove_writer()  # Pending buffered data will not be written
        self._loop.call_soon(self._call_connection_lost, exc)

    def _call_connection_lost(self, exc):
        """Close the connection.

        Informs the protocol through connection_lost() and clears
        pending buffers and closes the serial connection.
        """
        assert self._closing
        assert not self._has_writer
        assert not self._has_reader
        try:
            self._serial.flush()
        except (serial.SerialException if os.name == "nt" else termios.error):
            # ignore serial errors which may happen if the serial device was
            # hot-unplugged.
            pass
        try:
            self._protocol.connection_lost(exc)
        finally:
            self._write_buffer.clear()
            self._serial.close()
            self._serial = None
            self._protocol = None
            self._loop = None


async def create_serial_connection(loop, protocol_factory, *args, **kwargs):
    """Create a connection to a new serial port instance.

    This function is a coroutine which will try to establish the
    connection.

    The chronological order of the operation is:

    1. protocol_factory is called without arguments and must return
       a protocol instance.

    2. The protocol instance is tied to the transport

    3. This coroutine returns successfully with a (transport,
       protocol) pair.

    4. The connection_made() method of the protocol
       will be called at some point by the event loop.

    Note:  protocol_factory can be any kind of callable, not
    necessarily a class. For example, if you want to use a pre-created
    protocol instance, you can pass lambda: my_protocol.

    Any additional arguments will be forwarded to the Serial constructor.
    """
    serial_instance = serial.serial_for_url(*args, **kwargs)
    transport, protocol = await connection_for_serial(loop, protocol_factory, serial_instance)
    return transport, protocol


async def connection_for_serial(loop, protocol_factory, serial_instance):
    """Create a connection to the given serial port instance.

    This function is a coroutine which will try to establish the
    connection.

    The chronological order of the operation is:

    1. protocol_factory is called without arguments and must return
       a protocol instance.

    2. The protocol instance is tied to the transport

    3. This coroutine returns successfully with a (transport,
       protocol) pair.

    4. The connection_made() method of the protocol
       will be called at some point by the event loop.

    Note:  protocol_factory can be any kind of callable, not
    necessarily a class. For example, if you want to use a pre-created
    protocol instance, you can pass lambda: my_protocol.
    """
    protocol = protocol_factory()
    transport = SerialTransport(loop, protocol, serial_instance)
    return transport, protocol


async def open_serial_connection(*, loop=None, limit=None, **kwargs):
    """A wrapper for create_serial_connection() returning a (reader,
    writer) pair.

    The reader returned is a StreamReader instance; the writer is a
    StreamWriter instance.

    The arguments are all the usual arguments to Serial(). Additional
    optional keyword arguments are loop (to set the event loop instance
    to use) and limit (to set the buffer limit passed to the
    StreamReader.

    This function is a coroutine.
    """
    if loop is None:
        loop = asyncio.get_event_loop()
    if limit is None:
        limit = asyncio.streams._DEFAULT_LIMIT
    reader = asyncio.StreamReader(limit=limit, loop=loop)
    protocol = asyncio.StreamReaderProtocol(reader, loop=loop)
    transport, _ = await create_serial_connection(
        loop=loop,
        protocol_factory=lambda: protocol,
        **kwargs)
    writer = asyncio.StreamWriter(transport, protocol, reader, loop)
    return reader, writer



#!/bin/env python3

import click
import logging
import os
import tempfile
import struct
import sys
import csv
import binascii
import esptool

from .idrc_esp_tools.gen_esp32part import PartitionTable
from .idrc_esp_tools import analyze_nvs


logger = logging.getLogger('configurer')
logger.setLevel(logging.INFO)

BASE_DIR = os.path.dirname(__file__)
TMP_DIR = os.path.join(os.environ['HOME'], '.hwtester')

NVS_TARGET_NAMESPACE = 'ESP-MDF'
exit_code = 0


class SerialDumper:

    def __init__(self, usb_port, baudrate):
        self.usb_port = usb_port
        self.baudrate = baudrate

    def get_nvs(self, nvs_key):

        os.makedirs(TMP_DIR, exist_ok=True)

        with tempfile.NamedTemporaryFile(dir=TMP_DIR, suffix='.bin') as f:

            # dump partition table from a running node
            esptool.main([
                '-p',
                self.usb_port,
                '--after',
                'no_reset',
                'read_flash',
                '0x8000',
                '0x1000',
                f.name,
            ])
            logger.info('Dumped partition table')

            f.seek(0)
            partition_table = PartitionTable.from_binary(f.read())
            nvs_partition = partition_table.find_by_name('nvs')
            assert nvs_partition, 'NVS partition not found!'
            logger.info('NVS partition @%s len: %s', nvs_partition.offset, nvs_partition.size)

        with tempfile.NamedTemporaryFile(dir=TMP_DIR, suffix='.bin') as f:

            # dump NVS from a running node
            esptool.main([
                '-p',
                self.usb_port,
                'read_flash',
                str(nvs_partition.offset),
                str(nvs_partition.size),
                f.name,
            ])
            logger.info('Dumped NVS partition')

            # get CSV of NVS values
            f.seek(0)
            nvs_data = f.read()
            analyze_nvs.verify_nvs_size(nvs_data)
            analyze_nvs.parse_nvs_binary(nvs_data)
            for line in analyze_nvs.get_nvs_data(only_namespace=NVS_TARGET_NAMESPACE):
                k, data, entry_type, value = line.split(',')
                if k == nvs_key:
                    return binascii.unhexlify(value)


@click.command()
@click.option('-p', '--port', type=str, required=True, help='USB port to open serial to.')
@click.option('-b', '--baudrate', type=int, default=115200, help='Port baudrate.')
def main(
    port,
    baudrate,
):

    runner = SerialDumper(usb_port=port, baudrate=baudrate)
    data = runner.get_nvs('mconfig_data')

    (
        router_ssid,
        router_password,
        router_bssid,
        mesh_id,
        mesh_password,
        mesh_type,
        channel,
        channel_switch_disable,
        router_switch_disable,
        mqtt_uri,
        _,
    ) = struct.unpack(
        '32s' # char router_ssid[32];
        '64s' # char router_password[64];
        '6s' # uint8_t router_bssid[6];
        '6s' # uint8_t mesh_id[6];
        '64s' # char mesh_password[64];
        'B' # mwifi_node_type_t mesh_type;
        'B' # uint8_t channel;
        'B' # uint8_t channel_switch_disable;
        'B' # uint8_t router_switch_disable;
        '128s' # uint8_t custom[32 + 96];
        'H', # uint16_t whitelist_size;
        data[:306]
    )

    mesh_type = {
        0: 'idle',
        1: 'root',
        2: 'node',
        3: 'leaf',
        4: 'sta',
    }.get(mesh_type)

    print('router_ssid =', router_ssid.rstrip(b'\x00').decode())
    print('router_password =', router_password.rstrip(b'\x00').decode())
    print('router_bssid =', binascii.hexlify(router_bssid).decode())
    print('mesh_id =', binascii.hexlify(mesh_id).decode())
    print('mesh_password =', mesh_password.rstrip(b'\x00').decode())
    print('mesh_type =', mesh_type)
    print('channel =', channel)
    print('channel_switch_disable =', channel_switch_disable)
    print('router_switch_disable =', router_switch_disable)
    print('mqtt_uri =', mqtt_uri.rstrip(b'\x00').decode())


def run_dumper():
    handler = logging.StreamHandler()
    formatter = logging.Formatter('%(levelname)-10s %(name)s - %(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    main()


if __name__ == '__main__':
    run_dumper()



#!/bin/env python3

import click
from serial import Serial
from umodbus.server.serial import get_server
from umodbus.server.serial.rtu import RTUServer


def read_register(slave_id, function_code, address):
    return 42


@click.command()
@click.option("-p", "--port", type=str, required=True, help="Serial port to bind to.")
def main(port):
    s = Serial(port)
    s.timeout = 10
    app = get_server(RTUServer, s)
    app.route_map.add_rule(read_register, slave_ids=list(range(1, 243)), function_codes=[4], addresses=[9, 10, 11])
    try:
        app.serve_forever()
    finally:
        app.shutdown()


def run_modbus_server():
    main()


if __name__ == "__main__":
    run_modbus_server()



#!/usr/bin/env python
#
# esp-idf NVS partition generation tool. Tool helps in generating NVS-compatible
# partition binary, with key-value pair entries provided via a CSV file.
#
# Copyright 2018 Espressif Systems (Shanghai) PTE LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import argparse
import array
import binascii
import codecs
import datetime
import distutils.dir_util
import os
import random
import struct
import sys
import zlib
from builtins import bytes, int, range
from io import open

from itertools import zip_longest


VERSION1_PRINT = 'V1 - Multipage Blob Support Disabled'
VERSION2_PRINT = 'V2 - Multipage Blob Support Enabled'


def reverse_hexbytes(addr_tmp):
    addr = []
    reversed_bytes = ''
    for i in range(0, len(addr_tmp), 2):
        addr.append(addr_tmp[i:i + 2])
    reversed_bytes = ''.join(reversed(addr))

    return reversed_bytes


""" Class for standard NVS page structure """


class Page(object):
    PAGE_PARAMS = {
        'max_size': 4096,
        'max_old_blob_size': 1984,
        'max_new_blob_size': 4000,
        'max_entries': 126
    }

    # Item type codes
    U8 = 0x01
    I8 = 0x11
    U16 = 0x02
    I16 = 0x12
    U32 = 0x04
    I32 = 0x14
    U64 = 0x08
    I64 = 0x18
    SZ = 0x21
    BLOB = 0x41
    BLOB_DATA = 0x42
    BLOB_IDX = 0x48

    # Few Page constants
    HEADER_SIZE = 32
    BITMAPARRAY_OFFSET = 32
    BITMAPARRAY_SIZE_IN_BYTES = 32
    FIRST_ENTRY_OFFSET = 64
    SINGLE_ENTRY_SIZE = 32
    CHUNK_ANY = 0xFF
    ACTIVE = 0xFFFFFFFE
    FULL = 0xFFFFFFFC
    VERSION1 = 0xFF
    VERSION2 = 0xFE

    def __init__(self, page_num, version, is_rsrv_page=False):
        self.entry_num = 0
        self.bitmap_array = array.array('B')
        self.version = version
        self.page_buf = bytearray(b'\xff') * Page.PAGE_PARAMS['max_size']
        if not is_rsrv_page:
            self.bitmap_array = self.create_bitmap_array()
            self.set_header(page_num, version)

    def set_header(self, page_num, version):
        # set page state to active
        page_header = bytearray(b'\xff') * 32
        page_state_active_seq = Page.ACTIVE
        struct.pack_into('<I', page_header, 0,  page_state_active_seq)
        # set page sequence number
        struct.pack_into('<I', page_header, 4, page_num)
        # set version
        if version == Page.VERSION2:
            page_header[8] = Page.VERSION2
        elif version == Page.VERSION1:
            page_header[8] = Page.VERSION1
        # set header's CRC
        crc_data = bytes(page_header[4:28])
        crc = zlib.crc32(crc_data, 0xFFFFFFFF)
        struct.pack_into('<I', page_header, 28, crc & 0xFFFFFFFF)
        self.page_buf[0:len(page_header)] = page_header

    def create_bitmap_array(self):
        bitarray = array.array('B')
        charsize = 32  # bitmaparray has 256 bits, hence 32 bytes
        fill = 255  # Fill all 8 bits with 1's
        bitarray.extend((fill,) * charsize)
        return bitarray

    def write_bitmaparray(self):
        bitnum = self.entry_num * 2
        byte_idx = bitnum // 8  # Find byte index in the array
        bit_offset = bitnum & 7  # Find bit offset in given byte index
        mask = ~(1 << bit_offset)
        self.bitmap_array[byte_idx] &= mask
        start_idx = Page.BITMAPARRAY_OFFSET
        end_idx = Page.BITMAPARRAY_OFFSET + Page.BITMAPARRAY_SIZE_IN_BYTES
        self.page_buf[start_idx:end_idx] = self.bitmap_array

    def write_entry_to_buf(self, data, entrycount,nvs_obj):
        data_offset = Page.FIRST_ENTRY_OFFSET + (Page.SINGLE_ENTRY_SIZE * self.entry_num)
        start_idx = data_offset
        end_idx = data_offset + len(data)
        self.page_buf[start_idx:end_idx] = data

        # Set bitmap array for entries in current page
        for i in range(0, entrycount):
            self.write_bitmaparray()
            self.entry_num += 1

    def set_crc_header(self, entry_struct):
        crc_data = bytearray(b'28')
        crc_data[0:4] = entry_struct[0:4]
        crc_data[4:28] = entry_struct[8:32]
        crc_data = bytes(crc_data)
        crc = zlib.crc32(crc_data, 0xFFFFFFFF)
        struct.pack_into('<I', entry_struct, 4, crc & 0xFFFFFFFF)
        return entry_struct

    def write_varlen_binary_data(self, entry_struct, ns_index, key, data, data_size, total_entry_count, encoding, nvs_obj):
        chunk_start = 0
        chunk_count = 0
        chunk_index = Page.CHUNK_ANY
        offset = 0
        remaining_size = data_size
        tailroom = None

        while True:
            chunk_size = 0

            # Get the size available in current page
            tailroom = (Page.PAGE_PARAMS['max_entries'] - self.entry_num - 1) * Page.SINGLE_ENTRY_SIZE
            assert tailroom >= 0, 'Page overflow!!'

            # Split the binary data into two and store a chunk of available size onto curr page
            if tailroom < remaining_size:
                chunk_size = tailroom
            else:
                chunk_size = remaining_size

            remaining_size = remaining_size - chunk_size

            # Change type of data to BLOB_DATA
            entry_struct[1] = Page.BLOB_DATA

            # Calculate no. of entries data chunk will require
            datachunk_rounded_size = (chunk_size + 31) & ~31
            datachunk_entry_count = datachunk_rounded_size // 32
            datachunk_total_entry_count = datachunk_entry_count + 1  # +1 for the entry header

            # Set Span
            entry_struct[2] = datachunk_total_entry_count

            # Update the chunkIndex
            chunk_index = chunk_start + chunk_count
            entry_struct[3] = chunk_index

            # Set data chunk
            data_chunk = data[offset:offset + chunk_size]

            # Compute CRC of data chunk
            struct.pack_into('<H', entry_struct, 24, chunk_size)

            if type(data) != bytes:
                data_chunk = bytes(data_chunk, encoding='utf8')

            crc = zlib.crc32(data_chunk, 0xFFFFFFFF)
            struct.pack_into('<I', entry_struct, 28, crc & 0xFFFFFFFF)

            # compute crc of entry header
            entry_struct = self.set_crc_header(entry_struct)

            # write entry header
            self.write_entry_to_buf(entry_struct, 1,nvs_obj)
            # write actual data
            self.write_entry_to_buf(data_chunk, datachunk_entry_count,nvs_obj)

            chunk_count = chunk_count + 1

            if remaining_size or (tailroom - chunk_size) < Page.SINGLE_ENTRY_SIZE:
                nvs_obj.create_new_page()
                self = nvs_obj.cur_page

            offset = offset + chunk_size

            # All chunks are stored, now store the index
            if not remaining_size:
                # Initialise data field to 0xff
                data_array = bytearray(b'\xff') * 8
                entry_struct[24:32] = data_array

                # change type of data to BLOB_IDX
                entry_struct[1] = Page.BLOB_IDX

                # Set Span
                entry_struct[2] = 1

                # Update the chunkIndex
                chunk_index = Page.CHUNK_ANY
                entry_struct[3] = chunk_index

                struct.pack_into('<I', entry_struct, 24, data_size)
                entry_struct[28] = chunk_count
                entry_struct[29] = chunk_start

                # compute crc of entry header
                entry_struct = self.set_crc_header(entry_struct)

                # write last entry
                self.write_entry_to_buf(entry_struct, 1,nvs_obj)
                break

        return entry_struct

    def write_single_page_entry(self, entry_struct, data, datalen, data_entry_count, nvs_obj):
        # compute CRC of data
        struct.pack_into('<H', entry_struct, 24, datalen)

        if type(data) != bytes:
            data = bytes(data, encoding='utf8')

        crc = zlib.crc32(data, 0xFFFFFFFF)
        struct.pack_into('<I', entry_struct, 28, crc & 0xFFFFFFFF)

        # compute crc of entry header
        entry_struct = self.set_crc_header(entry_struct)

        # write entry header
        self.write_entry_to_buf(entry_struct, 1, nvs_obj)
        # write actual data
        self.write_entry_to_buf(data, data_entry_count, nvs_obj)

    """
    Low-level function to write variable length data into page buffer. Data should be formatted
    according to encoding specified.
    """
    def write_varlen_data(self, key, data, encoding, ns_index,nvs_obj):
        # Set size of data
        datalen = len(data)

        if datalen > Page.PAGE_PARAMS['max_old_blob_size']:
            if self.version == Page.VERSION1:
                raise InputError(' Input File: Size (%d) exceeds max allowed length `%s` bytes for key `%s`.'
                                 % (datalen, Page.PAGE_PARAMS['max_old_blob_size'], key))
            else:
                if encoding == 'string':
                    raise InputError(' Input File: Size (%d) exceeds max allowed length `%s` bytes for key `%s`.'
                                     % (datalen, Page.PAGE_PARAMS['max_old_blob_size'], key))

        # Calculate no. of entries data will require
        rounded_size = (datalen + 31) & ~31
        data_entry_count = rounded_size // 32
        total_entry_count = data_entry_count + 1  # +1 for the entry header

        # Check if page is already full and new page is needed to be created right away
        if self.entry_num >= Page.PAGE_PARAMS['max_entries']:
            raise PageFullError()
        elif (self.entry_num + total_entry_count) >= Page.PAGE_PARAMS['max_entries']:
            if not (self.version == Page.VERSION2 and encoding in ['hex2bin', 'binary', 'base64']):
                raise PageFullError()

        # Entry header
        entry_struct = bytearray(b'\xff') * 32
        # Set Namespace Index
        entry_struct[0] = ns_index
        # Set Span
        if self.version == Page.VERSION2:
            if encoding == 'string':
                entry_struct[2] = data_entry_count + 1
            # Set Chunk Index
            chunk_index = Page.CHUNK_ANY
            entry_struct[3] = chunk_index
        else:
            entry_struct[2] = data_entry_count + 1

        # set key
        key_array = b'\x00' * 16
        entry_struct[8:24] = key_array
        entry_struct[8:8 + len(key)] = key.encode()

        # set Type
        if encoding == 'string':
            entry_struct[1] = Page.SZ
        elif encoding in ['hex2bin', 'binary', 'base64']:
            entry_struct[1] = Page.BLOB

        if self.version == Page.VERSION2 and (encoding in ['hex2bin', 'binary', 'base64']):
            entry_struct = self.write_varlen_binary_data(entry_struct,ns_index,key,data,
                                                         datalen,total_entry_count, encoding, nvs_obj)
        else:
            self.write_single_page_entry(entry_struct, data, datalen, data_entry_count, nvs_obj)

    """ Low-level function to write data of primitive type into page buffer. """
    def write_primitive_data(self, key, data, encoding, ns_index,nvs_obj):
        # Check if entry exceeds max number of entries allowed per page
        if self.entry_num >= Page.PAGE_PARAMS['max_entries']:
            raise PageFullError()

        entry_struct = bytearray(b'\xff') * 32
        entry_struct[0] = ns_index  # namespace index
        entry_struct[2] = 0x01  # Span
        chunk_index = Page.CHUNK_ANY
        entry_struct[3] = chunk_index

        # write key
        key_array = b'\x00' * 16
        entry_struct[8:24] = key_array
        entry_struct[8:8 + len(key)] = key.encode()

        if encoding == 'u8':
            entry_struct[1] = Page.U8
            struct.pack_into('<B', entry_struct, 24, data)
        elif encoding == 'i8':
            entry_struct[1] = Page.I8
            struct.pack_into('<b', entry_struct, 24, data)
        elif encoding == 'u16':
            entry_struct[1] = Page.U16
            struct.pack_into('<H', entry_struct, 24, data)
        elif encoding == 'i16':
            entry_struct[1] = Page.I16
            struct.pack_into('<h', entry_struct, 24, data)
        elif encoding == 'u32':
            entry_struct[1] = Page.U32
            struct.pack_into('<I', entry_struct, 24, data)
        elif encoding == 'i32':
            entry_struct[1] = Page.I32
            struct.pack_into('<i', entry_struct, 24, data)
        elif encoding == 'u64':
            entry_struct[1] = Page.U64
            struct.pack_into('<Q', entry_struct, 24, data)
        elif encoding == 'i64':
            entry_struct[1] = Page.I64
            struct.pack_into('<q', entry_struct, 24, data)

        # Compute CRC
        crc_data = bytearray(b'28')
        crc_data[0:4] = entry_struct[0:4]
        crc_data[4:28] = entry_struct[8:32]
        crc_data = bytes(crc_data)
        crc = zlib.crc32(crc_data, 0xFFFFFFFF)
        struct.pack_into('<I', entry_struct, 4, crc & 0xFFFFFFFF)

        # write to file
        self.write_entry_to_buf(entry_struct, 1,nvs_obj)

    """ Get page buffer data of a given page """
    def get_data(self):
        return self.page_buf


"""
NVS class encapsulates all NVS specific operations to create a binary with given key-value pairs.
Binary can later be flashed onto device via a flashing utility.
"""


class NVS(object):
    def __init__(self, fout, input_size, version, encrypt=False, key_input=None):
        self.size = input_size
        self.encrypt = encrypt
        self.encr_key = None
        self.namespace_idx = 0
        self.page_num = -1
        self.pages = []
        self.version = version
        self.fout = fout
        if self.encrypt:
            self.encr_key = key_input
        self.cur_page = self.create_new_page(version)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None and exc_value is None:
            # Create pages for remaining available size
            while True:
                try:
                    self.create_new_page()
                except InsufficientSizeError:
                    self.size = None
                    # Creating the last reserved page
                    self.create_new_page(is_rsrv_page=True)
                    break
            result = self.get_binary_data()
            self.fout.write(result)

    def create_new_page(self, version=None, is_rsrv_page=False):
        # Set previous page state to FULL before creating new page
        if self.pages:
            curr_page_state = struct.unpack('<I', self.cur_page.page_buf[0:4])[0]
            if curr_page_state == Page.ACTIVE:
                page_state_full_seq = Page.FULL
                struct.pack_into('<I', self.cur_page.page_buf, 0, page_state_full_seq)
        # Set version for NVS binary generated
        version = self.version
        # Update available size as each page is created
        if self.size == 0:
            raise InsufficientSizeError('Error: Size parameter is less than the size of data in csv.Please increase size.')
        if not is_rsrv_page:
            self.size = self.size - Page.PAGE_PARAMS['max_size']
        self.page_num += 1
        # Set version for each page and page header
        new_page = Page(self.page_num, version, is_rsrv_page)
        self.pages.append(new_page)
        self.cur_page = new_page
        return new_page

    """
    Write namespace entry and subsequently increase namespace count so that all upcoming entries
    will be mapped to a new namespace.
    """
    def write_namespace(self, key):
        self.namespace_idx += 1
        try:
            self.cur_page.write_primitive_data(key, self.namespace_idx, 'u8', 0,self)
        except PageFullError:
            new_page = self.create_new_page()
            new_page.write_primitive_data(key, self.namespace_idx, 'u8', 0,self)

    """
    Write key-value pair. Function accepts value in the form of ascii character and converts
    it into appropriate format before calling Page class's functions to write entry into NVS format.
    Function handles PageFullError and creates a new page and re-invokes the function on a new page.
    We don't have to guard re-invocation with try-except since no entry can span multiple pages.
    """
    def write_entry(self, key, value, encoding):
        if encoding == 'hex2bin':
            value = value.strip()
            if len(value) % 2 != 0:
                raise InputError('%s: Invalid data length. Should be multiple of 2.' % key)
            value = binascii.a2b_hex(value)

        if encoding == 'base64':
            value = binascii.a2b_base64(value)

        if encoding == 'string':
            if type(value) == bytes:
                value = value.decode()
            value += '\0'

        encoding = encoding.lower()
        varlen_encodings = ['string', 'binary', 'hex2bin', 'base64']
        primitive_encodings = ['u8', 'i8', 'u16', 'i16', 'u32', 'i32', 'u64', 'i64']

        if encoding in varlen_encodings:
            try:
                self.cur_page.write_varlen_data(key, value, encoding, self.namespace_idx,self)
            except PageFullError:
                new_page = self.create_new_page()
                new_page.write_varlen_data(key, value, encoding, self.namespace_idx,self)
        elif encoding in primitive_encodings:
            try:
                self.cur_page.write_primitive_data(key, int(value), encoding, self.namespace_idx,self)
            except PageFullError:
                new_page = self.create_new_page()
                new_page.write_primitive_data(key, int(value), encoding, self.namespace_idx,self)
        else:
            raise InputError('%s: Unsupported encoding' % encoding)

    """ Return accumulated data of all pages """
    def get_binary_data(self):
        data = bytearray()
        for page in self.pages:
            data += page.get_data()
        return data


class PageFullError(RuntimeError):
    """
    Represents error when current page doesn't have sufficient entries left
    to accommodate current request
    """
    def __init__(self):
        super(PageFullError, self).__init__()


class InputError(RuntimeError):
    """
    Represents error on the input
    """
    def __init__(self, e):
        print('\nError:')
        super(InputError, self).__init__(e)


class InsufficientSizeError(RuntimeError):
    """
    Represents error when NVS Partition size given is insufficient
    to accomodate the data in the given csv file
    """
    def __init__(self, e):
        super(InsufficientSizeError, self).__init__(e)


def nvs_open(result_obj, input_size, version=None, is_encrypt=False, key=None):
    """ Wrapper to create and NVS class object. This object can later be used to set key-value pairs

    :param result_obj: File/Stream object to dump resultant binary. If data is to be dumped into memory, one way is to use BytesIO object
    :param input_size: Size of Partition
    :return: NVS class instance
    """
    return NVS(result_obj, input_size, version, encrypt=is_encrypt, key_input=key)


def write_entry(nvs_instance, key, datatype, encoding, value):
    """ Wrapper to set key-value pair in NVS format

    :param nvs_instance: Instance of an NVS class returned by nvs_open()
    :param key: Key of the data
    :param datatype: Data type. Valid values are "file", "data" and "namespace"
    :param encoding: Data encoding. Valid values are "u8", "i8", "u16", "i16", "u32", "i32", "u64", "i64", "string", "binary", "hex2bin" and "base64"
    :param value: Data value in ascii encoded string format for "data" datatype and filepath for "file" datatype
    :return: None
    """

    if datatype == 'file':
        abs_file_path = value
        if os.path.isabs(value) is False:
            script_dir = os.getcwd()
            abs_file_path = os.path.join(script_dir, value)

        with open(abs_file_path, 'rb') as f:
            value = f.read()

    if datatype == 'namespace':
        nvs_instance.write_namespace(key)
    else:
        nvs_instance.write_entry(key, value, encoding)


def nvs_close(nvs_instance):
    """ Wrapper to finish writing to NVS and write data to file/stream object provided to nvs_open method

    :param nvs_instance: Instance of NVS class returned by nvs_open()
    :return: None
    """
    nvs_instance.__exit__(None, None, None)


def check_size(size):
    '''
    Checks for input partition size
    :param size: Input partition size
    '''
    try:
        # Set size
        input_size = int(size, 0)
        if input_size % 4096 != 0:
            sys.exit('Size of partition must be multiple of 4096')

        # Update size as a page needs to be reserved of size 4KB
        input_size = input_size - Page.PAGE_PARAMS['max_size']

        if input_size < (2 * Page.PAGE_PARAMS['max_size']):
            sys.exit('Minimum NVS partition size needed is 0x3000 bytes.')
        return input_size
    except Exception as e:
        print(e)
        sys.exit(0)


def set_target_filepath(outdir, filepath):
    '''
    Set target file path: <outdir>/<filepath>
    :param outdir: Target output dir to store files
    :param filepath: Path of target file
    '''
    bin_ext = '.bin'
    # Expand if tilde(~) provided in path
    outdir = os.path.expanduser(outdir)

    if filepath:
        key_file_name, ext = os.path.splitext(filepath)
        if not ext:
            filepath = key_file_name + bin_ext
        elif bin_ext not in ext:
            sys.exit('Error: `%s`. Only `%s` extension allowed.' % (filepath, bin_ext))

    # Create dir if does not exist
    if not (os.path.isdir(outdir)):
        distutils.dir_util.mkpath(outdir)

    filedir, filename = os.path.split(filepath)
    filedir = os.path.join(outdir,filedir,'')
    if filedir and not os.path.isdir(filedir):
        distutils.dir_util.mkpath(filedir)

    if os.path.isabs(filepath):
        if not outdir == os.getcwd():
            print('\nWarning: `%s` \n\t==> absolute path given so outdir is ignored for this file.' % filepath)
        # Set to empty as outdir is ignored here
        outdir = ''

    # Set full path - outdir + filename
    filepath = os.path.join(outdir, '') + filepath

    return outdir, filepath


def generate(args, is_encr_enabled=False, encr_key=None):
    '''
    Generate NVS Partition
    :param args: Command line arguments given
    :param is_encr_enabled: Encryption enabled/disabled
    :param encr_key: Key to encrypt NVS partition
    '''
    is_dir_new = False
    bin_ext = '.bin'

    input_size = check_size(args.size)
    if args.version == 1:
        args.version = Page.VERSION1
    elif args.version == 2:
        args.version = Page.VERSION2

    # Check if key file has .bin extension
    filename, ext = os.path.splitext(args.output)
    if bin_ext not in ext:
        sys.exit('Error: `%s`. Only `.bin` extension allowed.' % args.output)
    args.outdir, args.output = set_target_filepath(args.outdir, args.output)

    input_file = open(args.input, 'rt', encoding='utf8')
    output_file = open(args.output, 'wb')

    with open(args.input, 'rt', encoding='utf8') as input_file,\
            open(args.output, 'wb') as output_file,\
            nvs_open(output_file, input_size, args.version, is_encrypt=is_encr_enabled, key=encr_key) as nvs_obj:

        if nvs_obj.version == Page.VERSION1:
            version_set = VERSION1_PRINT
        else:
            version_set = VERSION2_PRINT

        print('\nCreating NVS binary with version:', version_set)

        line = input_file.readline().strip()

        # Comments are skipped
        while line.startswith('#'):
            line = input_file.readline().strip()
        if not isinstance(line, str):
            line = line.encode('utf-8')

        header = line.split(',')

        while True:
            line = input_file.readline().strip()
            if not isinstance(line, str):
                line = line.encode('utf-8')

            value = line.split(',')
            if len(value) == 1 and '' in value:
                break

            data = dict(zip_longest(header, value))

            try:
                # Check key length
                if len(data['key']) > 15:
                    raise InputError('Length of key `{}` should be <= 15 characters.'.format(data['key']))
                write_entry(nvs_obj, data['key'], data['type'], data['encoding'], data['value'])
            except InputError as e:
                print(e)
                filedir, filename = os.path.split(args.output)
                if filename:
                    print('\nWarning: NVS binary not created...')
                    os.remove(args.output)
                if is_dir_new and not filedir == os.getcwd():
                        print('\nWarning: Output dir not created...')
                        os.rmdir(filedir)
                sys.exit(-2)

    print('\nCreated NVS binary: ===>', args.output)


def main():
    parser = argparse.ArgumentParser()
    subparser = parser.add_subparsers()
    parser_gen = subparser.add_parser('generate',
                                      help='Generate NVS partition',
                                      formatter_class=argparse.RawTextHelpFormatter)
    parser_gen.set_defaults(func=generate)
    parser_gen.add_argument('input',
                            default=None,
                            help='Path to CSV file to parse')
    parser_gen.add_argument('output',
                            default=None,
                            help='Path to output NVS binary file')
    parser_gen.add_argument('size',
                            default=None,
                            help='Size of NVS partition in bytes\
                            \n(must be multiple of 4096)')
    parser_gen.add_argument('--version',
                            choices=[1,2],
                            default=2,
                            type=int,
                            help='''Set multipage blob version.\
                            \nVersion 1 - Multipage blob support disabled.\
                            \nVersion 2 - Multipage blob support enabled.\
                            \nDefault: Version 2''')
    parser_gen.add_argument('--outdir',
                            default=os.getcwd(),
                            help='Output directory to store files created\
                            \n(Default: current directory)')
    args = parser.parse_args()
    args.func(args)


if __name__ == '__main__':
    main()



"""Test Main Module

.. todo::

    Add tests for I2C communication, as a way to test the main board without I/O
    board attached to it. The tests provided here assume physical relays are in 
    place, since it uses the Raspberry Pi GPIOs to assert the relay state. The 
    I2C bus of the Raspberry Pi can be used for this purpose.
"""


import asyncio
import logging
import sys

from teramesh_hardware_tester.tester import getch

try:
    import RPi.GPIO as GPIO
except ImportError:
    GPIO = None

logger = logging.getLogger('test')


test_calibration_temperature = None
test_calibration_humidity = None

#: Raspberry Pi common GPIO output
RELAY_COMMON_GPIO = 5
#: Raspberry Pi GPIO input
RELAY_GPIO_MAP = {
    1: 4,
    2: 17,
    3: 27,
    4: 22,
    5: 23,
    6: 24,
    7: 25,
}


async def setup(AT):
    global test_calibration_temperature
    global test_calibration_humidity
    try:
        from teramesh_hardware_tester.idrc_esp_tools.sht85 import single_shot
        test_calibration_temperature, test_calibration_humidity = single_shot()
    except Exception as e:
        logger.warning('SHT85 failed to read: %s', e)


async def test_comms(AT):
    """Test Communications
    
    This function internally uses the ``TESTCOMMS`` cat command.
    """

    await AT.TESTCOMMS.execute()
    await AT.TESTCOMMS.expect_ok_or_error()


async def test_shtc3(AT):
    """Test SHTC3
    
    This function internally uses the ``TESTSHTC3`` cat command.
    """

    global test_calibration_temperature
    global test_calibration_humidity
    await AT.TESTSHTC3.execute()
    result = await AT.TESTSHTC3.expect_reply()
    temperature, humidity = result.strip().split(',')
    logger.info('SHTC3 temperature %s', temperature)
    logger.info('SHTC3 humidity %s', humidity)
    # FIXME: TODO: RPi could be connected to a calibration source and compare these to some ground truth
    assert 10 < float(temperature) < 50, 'SHTC3 returned erroneous temperature'
    assert 0 < float(humidity) < 100, 'SHTC3 returned erroneous humidity'
    if test_calibration_temperature is not None:
        logger.info('SHTC3 temperature offset: %f', test_calibration_temperature - float(temperature))
    if test_calibration_humidity is not None:
        logger.info('SHTC3 humidity offset: %f', test_calibration_humidity - float(humidity))


async def test_ds18b20(AT):
    """Test DS18B20

    This function internally uses the ``TESTDS18B20`` cat command.
    """

    await AT.TESTDS18B20.execute()
    temperature = await AT.TESTDS18B20.expect_reply()
    logger.info('DS18B20 temperature %s', temperature)
    # FIXME: TODO: RPi could be connected to a calibration source and compare these to some ground truth
    assert 10 < float(temperature) < 50, 'DS18B20 returned erroneous temperature'


async def test_modbus(AT):
    """Test Modbus

    This function internally uses the ``TESTMODBUS`` cat command.
    """

    await AT.TESTMODBUS.set(
        # slave address: raw value
        1,
        # baud rate: raw value (ex: 1200 or 2400 or 4800 or 9600 or 14400 or ...)
        9600,
        # parity: [1 = even][2 = odd][3 = none] (cf. protocol)
        3,
        # function code: [1 = discret input][2 = coil][3 = input register][4 = holding register] (cf protocol)
        3,
        # register address: raw value
        10,
        # word/byte order: [1 = big endian][2 = little endian] (cf protocol)
        2,
        2,
        # encoding: [1 = bool][2 = float][3 = double][4 = int8][5 = int16][6 = int32][7 = int64][8 = uint8][9 = uint16][10 = uint32][11 = uint64] (cf protocol)
        9,
    )
    result = await AT.TESTMODBUS.expect_reply()
    assert int(result) == 42, 'Modbus returned erroneous result'


async def test_buzzer(AT):
    """Test buzzer
    
    This function internally uses the ``TESTBUZZ`` cat command.
    """

    await AT.TESTBUZZ.execute()
    await AT.TESTBUZZ.expect_ok_or_error()
    logger.critical('Did the buzzer buzz (y/n)?')
    if sys.stdin.isatty():
        assert getch().lower() == 'y', 'Buzzer did not buzz'


async def test_led(AT):
    """Test LED display
    
    This function internally uses the ``TESTLED`` cat command.
    """

    await AT.TESTLED.execute()
    await AT.TESTLED.expect_ok_or_error()
    logger.critical('Did the LED screen do the thing (y/n)?')
    if sys.stdin.isatty():
        assert getch().lower() == 'y', 'LED did not do the thing'


async def test_pir(AT):
    """Test PIR
    
    This function internally uses the ``TESTPIR`` cat command.
    """

    logger.critical('Please WAVE AT PIR!')
    await AT.TESTPIR.execute()
    result = await AT.TESTPIR.expect_reply(timeout=10)
    assert int(result) == 1, 'PIR did not detect waving'


async def test_btn(AT):
    """Test button
    
    This function internally uses the ``TESTBTN`` cat command.
    """

    # TESTBTN handler waits for 5 seconds
    await AT.TESTBTN.execute()
    logger.critical('Please PRESS DEV KEY!')
    await AT.TESTBTN.expect_ok_or_error(timeout=15)


def relays_setup():
    if GPIO is None:
        assert False, 'RPi GPIO module is not installed!'
    GPIO.setmode(GPIO.BCM)
    GPIO.setup(RELAY_COMMON_GPIO, GPIO.OUT)
    GPIO.output(RELAY_COMMON_GPIO, GPIO.HIGH)
    for relay_pin in RELAY_GPIO_MAP.values():
        GPIO.setup(relay_pin, GPIO.IN, pull_up_down=GPIO.PUD_DOWN)

async def _test_relay(AT, relay_no: int):
    """Generic Test Relay

    This function internally uses the ``TESTRELAY`` cat command.

    This function first turns off the relay, then turns it on, and finally 
    turns it off again. At every step it checks through the Raspberry Pi GPIOs 
    if the relay state is the expected one.

    ``RELAY_COMMON_GPIO`` is used to produce an output signal, and the 
    ``RELAY_GPIO_MAP`` inputs are connected to it through the relays. This way,
    the program can assert if the relay is on or off.

    :param AT:
    :param relay_no: Relay number, from 1 to 7
    """

    relays_setup()
    try:
        await AT.TESTRELAY0.execute()
        try:
            await AT.TESTRELAY0.expect_ok_or_error()
        except Exception:
            pass
        relay_pin = RELAY_GPIO_MAP[relay_no]
        await getattr(AT, f'TESTRELAY{relay_no}').execute()
        await getattr(AT, f'TESTRELAY{relay_no}').expect_ok_or_error()
        await asyncio.sleep(0.1)
        assert GPIO.input(relay_pin), f'Relay {relay_no} did not open'
        await AT.TESTRELAY0.execute()
        await AT.TESTRELAY0.expect_ok_or_error()
        await asyncio.sleep(0.1)
        assert not GPIO.input(relay_pin), f'Relay {relay_no} did not close'
    finally:
        await AT.TESTRELAY0.execute()
        GPIO.cleanup()

async def test_relay1(AT):
    await _test_relay(AT, 1)

async def test_relay2(AT):
    await _test_relay(AT, 2)

async def test_relay3(AT):
    await _test_relay(AT, 3)

async def test_relay4(AT):
    await _test_relay(AT, 4)

async def test_relay5(AT):
    await _test_relay(AT, 5)

async def test_relay6(AT):
    await _test_relay(AT, 6)

async def test_relay7(AT):
    await _test_relay(AT, 7)

async def test_relays_4_and_6(AT):
    relays_setup()
    try:
        await AT.TESTRELAY0.execute()
        try:
            await AT.TESTRELAY0.expect_ok_or_error()
        except Exception:
            pass
        # turn on relay 4
        await AT.TESTRELAY4.execute()
        await AT.TESTRELAY4.expect_ok_or_error()
        await asyncio.sleep(0.1)
        assert GPIO.input(RELAY_GPIO_MAP[4]), 'Relay 4 did not open'
        # turn on relay 6
        await AT.TESTRELAY6.execute()
        await AT.TESTRELAY6.expect_ok_or_error()
        await asyncio.sleep(0.1)
        assert GPIO.input(RELAY_GPIO_MAP[6]), 'Relay 6 did not open'
        # after that relay 4 should turn off
        await asyncio.sleep(0.1)
        assert not GPIO.input(RELAY_GPIO_MAP[4]), 'Relay 4 did not close'
    finally:
        await AT.TESTRELAY0.execute()
        GPIO.cleanup()




import asyncio
import logging
import time


logger = logging.getLogger('at')


STREAM_HANDLERS = [
    # {
    #     'queue': queue that receives the reply,
    #     'prefixes': [list of expected prefixes for the messages],
    #     'started': ms timestamp when this handler created,
    #     'timeout': ms within which this handler expires,
    # },
]
STREAM_LOCK = asyncio.Lock()


def now():
    return time.time()


class ATCommandError(Exception):
    pass

class ATCommandTimeout(Exception):
    pass


class Command:

    def __init__(self, command, w, timeout=None):
        self.command = command
        self.w = w
        self.timeout = timeout or 0.5

    async def __write_line(self, line):
        async with STREAM_LOCK:
            logger.debug('> %s', line)
            self.w.write(f'{line}\r\n'.encode())

    async def simple(self, param):
        # ATXXX0
        await self.__write_line(f'AT{self.command}{param}')

    async def execute(self):
        # AT+XXX
        await self.__write_line(f'AT+{self.command}')

    async def read(self):
        # AT+XXX?
        await self.__write_line(f'AT+{self.command}?')

    async def set(self, *args, noquote=False):
        # AT+XXX=1,2,"3"
        params = []
        for p in args:
            if isinstance(p, str) and not noquote:
                params.append(f'"{p}"')
            elif isinstance(p, bytes) and not noquote:
                params.append(f'"{p.decode()}"')
            else:
                params.append(str(p))
        params = ','.join(params)
        await self.__write_line(f'AT+{self.command}={params}')

    async def expect_ok_or_error(self, timeout=None):
        global STREAM_HANDLERS
        start = now()
        queue = asyncio.Queue()
        STREAM_HANDLERS.append({
            'started': start,
            'timeout': timeout or self.timeout,
            'queue': queue,
            'prefixes': ['OK', 'ERROR'],
        })
        try:
            reply = await asyncio.wait_for(queue.get(), timeout or self.timeout)
        except asyncio.TimeoutError:
            raise ATCommandTimeout(f'{self.command} timed out after {now() - start} ms')
        if reply == 'ERROR':
            raise ATCommandError(f'{self.command} ERROR')

    async def expect_reply(self, timeout=None):
        # +XXX reply
        global STREAM_HANDLERS
        start = now()
        queue = asyncio.Queue()
        STREAM_HANDLERS.append({
            'started': start,
            'timeout': timeout or self.timeout,
            'queue': queue,
            'prefixes': ['ERROR', f'+{self.command}'],
        })
        try:
            reply = await asyncio.wait_for(queue.get(), timeout or self.timeout)
        except asyncio.TimeoutError:
            raise ATCommandTimeout(f'{self.command} timed out after {now() - start} sec')
        if reply.startswith(f'+{self.command}:'):
            return reply.replace(f'+{self.command}:', '').strip()
        raise ATCommandError(f'{self.command} ERROR')


class AT:
    # https://en.wikipedia.org/wiki/Hayes_command_set

    def __init__(self, r, w, command_timeouts=None, default_timeout=0.5, log=None):
        # stream is an object with `write` and `readline` methods
        self.r = r
        self.w = w
        self.command_timeouts = command_timeouts or {}
        self.default_timeout = default_timeout
        self.log = None
        if log:
            self.log = open(log, 'w+')
        asyncio.create_task(self.listen_stream())

    async def listen_stream(self):
        global STREAM_HANDLERS
        try:
            while True:
                line = await self.r.readline()
                try:
                    line = (line.replace(b'\xff', b'')).decode().strip()
                except Exception as e:
                    logger.error(str(e))
                    continue
                if not line:
                    continue
                logger.debug('< %s', line)
                if self.log:
                    self.log.write(line + '\n')
                    self.log.flush()
                # cleanup handlers
                STREAM_HANDLERS = [x for x in STREAM_HANDLERS if not x.get('delete')]
                for handler in STREAM_HANDLERS:
                    if handler.get('timeout') is not None:
                        if now() > handler['started'] + handler['timeout']:
                            # handler expired
                            handler['delete'] = True
                            continue
                        for prefix in handler['prefixes']:
                            if line.startswith(prefix):
                                handler['queue'].put_nowait(line)
        except asyncio.CancelledError:
            logger.info('Exiting AT stream listener...')

    def __getattr__(self, name):
        return Command(
            name,
            w=self.w,
            timeout=self.command_timeouts.get(name, self.default_timeout),
        )



#!/bin/env python3

import asyncio
import click
import csv
import esptool
import glob
import importlib
import logging
import os
import pathlib
import re
import subprocess
import sys
import tempfile
import termios
import tty
from serial.tools import list_ports

from .idrc_esp_tools import serial_asyncio, at
from .idrc_esp_tools.gen_esp32part import PartitionTable


logger = logging.getLogger('tester')
logger.setLevel(logging.INFO)

BASE_DIR = os.path.dirname(__file__)
TMP_DIR = os.path.join(os.environ['HOME'], '.hwtester')

exit_code = 0


def getch():
    def _getch():
        fd = sys.stdin.fileno()
        old_settings = termios.tcgetattr(fd)
        try:
            tty.setraw(fd)
            ch = sys.stdin.read(1)
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
        return ch
    return _getch()


class SerialTestRunner:

    def __init__(self, usb_port, baudrate):
        self.usb_port = usb_port
        self.baudrate = baudrate

    def set_nvs_testing_flag(self):

        os.makedirs(TMP_DIR, exist_ok=True)

        with tempfile.NamedTemporaryFile(dir=TMP_DIR, suffix='.bin') as f:

            # dump partition table from a running node
            esptool.main([
                '-p',
                self.usb_port,
                '--after',
                'no_reset',
                'read_flash',
                '0x8000',
                '0x1000',
                f.name,
            ])
            logger.info('Dumped partition table')

            f.seek(0)
            partition_table = PartitionTable.from_binary(f.read())
            test_partition = partition_table.find_by_name('test')
            assert test_partition, 'Test partition not found!'
            logger.info('Test partition @%s len: %s', test_partition.offset, test_partition.size)

        with tempfile.NamedTemporaryFile(dir=TMP_DIR, suffix='.bin') as f:

            # prepare test partition contents
            f.write(b'test')
            f.write(b'\x00' * (test_partition.size - 4))
            f.flush()

            # flash test partition and reboot the node
            esptool.main([
                '-p',
                self.usb_port,
                'write_flash',
                str(test_partition.offset),
                f.name,
            ])
            logger.info('Flashed test flag partition')

        # TODO: start testing
        logger.info('Ready to test!')

    @staticmethod
    def get_all_tests():
        for filename in glob.glob(os.path.join(BASE_DIR, 'tests/**/test*.py'), recursive=True):
            module = importlib.machinery.SourceFileLoader('test', filename).load_module()
            for func_name in dir(module):
                if func_name.startswith('test') and callable(getattr(module, func_name)):
                    yield {
                        'filename': os.path.basename(filename),
                        'func': func_name,
                    }

    @staticmethod
    def get_all_ports():
        for port in list_ports.comports():
            yield {
               'device': port.device,
               'name': port.name,
               'description': port.description,
               'hwid': port.hwid,
               'vid': port.vid,
               'pid': port.pid,
               'serial_number': port.serial_number,
               'location': port.location,
               'manufacturer': port.manufacturer,
               'product': port.product,
               'interface': port.interface,
            }

    async def run_test_cases(self, include, exclude, tests_list=True, tests_execute=True):
        global exit_code
        tests_passed = True
        include = include or []
        exclude = exclude or []
        if tests_execute:
            self.reader, self.writer = await serial_asyncio.open_serial_connection(
                url=self.usb_port,
                baudrate=self.baudrate,
            )
            self.at = at.AT(
                r=self.reader,
                w=self.writer,
                default_timeout=5,
            )
            try:
                await self.at.TEST.expect_reply(timeout=10)
            except Exception as e:
                logger.error('Hardware did not reply with +TEST acknowledgement')
                logger.error(e)
                exit_code = 1
                return exit_code
        if tests_list:
            logger.info('TESTS LIST:')
        for filename in glob.glob(os.path.join(BASE_DIR, 'tests/**/test*.py'), recursive=True):
            module = importlib.machinery.SourceFileLoader('test', filename).load_module()
            if tests_execute and hasattr(module, 'setup'):
                await module.setup(AT=self.at)
            for func_name in dir(module):
                if func_name.startswith('test') and callable(getattr(module, func_name)):
                    if tests_list:
                        logger.info('%s: %s', os.path.basename(filename), func_name)
                    if not tests_execute:
                        continue
                    skip = False
                    for incl in include:
                        if not re.match(incl, func_name):
                            skip = True
                    for excl in exclude:
                        if re.match(excl, func_name):
                            skip = True
                    if skip:
                        logger.warning('[?] %s skipped', func_name)
                        continue
                    try:
                        await getattr(module, func_name)(AT=self.at)
                    except Exception as e:
                        logger.error('[!] %s error %s', func_name, e)
                        tests_passed = False
                    else:
                        logger.info('[x] %s passed', func_name)
            if tests_execute and hasattr(module, 'teardown'):
                await module.teardown(AT=self.at)
        if not tests_execute:
            exit_code = 0
            return exit_code
        if tests_passed:
            logger.info('All tests passed, good job!')
            exit_code = 0
        else:
            logger.critical('NOTHING WORKS!!!111')
            exit_code = 1
        return exit_code


@click.command()
@click.option('-p', '--port', type=str, help='USB port to open serial to.')
@click.option('-b', '--baudrate', type=int, default=115200, help='Port baudrate.')
@click.option('-i', '--include', multiple=True, help='ONLY execute these test funcs (by func name, regex, multiple).')
@click.option('-x', '--exclude', multiple=True, help='DO NOT execute these test funcs (by func name, regex, multiple).')
@click.option('--tests-list/--no-tests-list', default=True)
@click.option('--tests-execute/--no-tests-execute', default=True)
def main(port, baudrate, include, exclude, tests_list, tests_execute):
    if tests_execute and not port:
        logger.error('--port is required when --tests-execute')
        sys.exit(1)
    runner = SerialTestRunner(usb_port=port, baudrate=baudrate)
    if tests_execute:
        runner.set_nvs_testing_flag()
    asyncio.run(runner.run_test_cases(
        include=include,
        exclude=exclude,
        tests_list=tests_list,
        tests_execute=tests_execute,
    ))

def run_tester():
    global exit_code

    handler = logging.StreamHandler()
    formatter = logging.Formatter('%(levelname)-10s %(name)s - %(message)s')
    handler.setFormatter(formatter)

    logger.addHandler(handler)

    test_logs = logging.getLogger('test')
    test_logs.setLevel(logging.DEBUG)
    test_logs.addHandler(handler)

    atlog = logging.getLogger('at')
    atlog.setLevel(logging.DEBUG)
    atlog.addHandler(logging.StreamHandler())

    # monkey-patch Click's sys.exit
    _exit = sys.exit
    sys.exit = lambda x: ...
    try:
        main()
    except Exception as e:
        logger.exception(e)
        exit_code = 1
    sys.exit = _exit

    sys.exit(exit_code)


if __name__ == '__main__':
    run_tester()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import io
import os
import sys
from shutil import rmtree

from setuptools import find_packages, setup, Command

# Package meta-data.
NAME = 'teramesh-hardware-tester'
DESCRIPTION = 'Scripts to test Teramesh hardware.'
URL = 'https://gitea.cluster.teramesh.cn/i-DRC/teramesh-hardware-tester'
EMAIL = 'michael.sayapin@i-drc.com'
AUTHOR = 'Michael Sayapin'
REQUIRES_PYTHON = '>=3.6.0'
VERSION = '0.7.2'

install_requires = [
    'pyserial>=3,<4',
    'esptool>=3.2,<4',
    'click>=8.0,<9',
    'asciimatics>=1.13,<2',
    'uModbus>=1,<2',
]

extras_require = {
    'rpi': [
        'RPi.GPIO ~= 0.7',
    ],
    'dev': [
        'ipython',
        'ipdb',
        'zest.releaser[recommended]',
    ],
    'ui': [
        'fastapi==0.70.1',
        'uvicorn==0.16.0',
        'jinja2==3.0.3',
    ],
}
extras_require['all'] = [v for s in extras_require.values() for v in s]

here = os.path.abspath(os.path.dirname(__file__))

try:
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

about = {}
if not VERSION:
    project_slug = NAME.lower().replace("-", "_").replace(" ", "_")
    with open(os.path.join(here, project_slug, '__version__.py')) as f:
        exec(f.read(), about)
else:
    about['__version__'] = VERSION


class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system('{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))

        self.status('Uploading the package to PyPI via Twine…')
        os.system('twine upload -r pypi dist/*')
        os.system('twine upload -r idrc dist/*')

        self.status('Pushing git tags…')
        os.system('git tag v{0}'.format(about['__version__']))
        os.system('git push --tags')

        sys.exit()


# Where the magic happens:
setup(
    name=NAME,
    version=about['__version__'],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'teramesh-hardware-test=teramesh_hardware_tester.tester:run_tester',
            'teramesh-modbus-test=teramesh_hardware_tester.modbuser:run_modbuser',
            'teramesh-configure=teramesh_hardware_tester.configurer:run_configurer',
            'teramesh-dump-config=teramesh_hardware_tester.dumper:run_dumper',
            'teramesh-modbus-server=teramesh_hardware_tester.modbus_server:run_modbus_server',
            'tht-ui=teramesh_hardware_tester.server:main',
        ],
    },
    install_requires=install_requires,
    extras_require=extras_require,
    include_package_data=True,
    license='Proprietary',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: Other/Proprietary License',
        'Natural Language :: English',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Testing :: Acceptance',
        'Topic :: System :: Hardware',
    ],
    cmdclass={
        'upload': UploadCommand,
    },
)
